@@ -1,10 +1,17 @@
 # Python -> C99, OpenCL, JS converter
 # created by Massimo Di Pierro
 # license: 2-clause BSD
-# requires meta (always), ezpyinline (if filter=ezpy) and numpy+pyopencl (if Device used)
+# requires meta (always) and numpy+pyopencl (if Device used)
 import logging
+import os
 import sys
+import imp
 import ast
+import uuid
+import threading
+import tempfile
+import shutil
+from meta.decompiler import decompile_func
 
 try:
     import numpy
@@ -13,28 +20,82 @@ try:
 except ImportError:
     HAVE_PYOPENCL = False
 
-try:
-    import ezpyinline
-    HAVE_EZPYINLINE = True
-except ImportError:
-    HAVE_EZPYINLINE = False
+__all__ = ['Compiler', 'C99Handler', 'JavaScriptHandler', 'Device']
 
-try:
-    from meta.decompiler import decompile_func
-except ImportError:
-    logging.error('must install "meta"')
-    sys.exit(1)
+locker = threading.Lock()
+
+C1 = '%s %s; if(!PyArg_ParseTuple(args, "%s", &%s)) return NULL;\n'
+C2 = '{"%(name)s", %(module_name)s_%(name)s, METH_VARARGS},\n'
+C3 = """
+static PyObject * %(module_name)s_%(name)s(PyObject *self, PyObject *args) {
+  int retval;
+  %(parsing_code)s
+  retval = %(name)s(%(args)s);
+  return Py_BuildValue("i", retval);
+} """
+C4 = """
+static PyMethodDef __%(name)s__%(module_name)s_Methods[] = {
+  %(funcs)s  {NULL, NULL}
+};
+DL_EXPORT(void) init%(module_name)s(void) {
+  Py_InitModule("%(module_name)s", __%(name)s__%(module_name)s_Methods);
+}
+"""
+
+def distutil_compile_and_import(module_name, code, build_dir=None):
+    from distutils.core import setup, Extension
+    from distutils.util import get_platform
+    if not build_dir:
+        build_dir = tempfile.mkdtemp()
+
+    # write code to file
+    file_name = os.path.join(build_dir, module_name + '.c')
+    open(file_name, 'w').write(code)
+    home_dir = os.getcwd()
+
+    # turn file into module
+    os.chdir(build_dir)
+    extension = Extension(module_name, [module_name + '.c'])
+    try:
+        locker.acquire()
+        setup(name=module_name,
+              version='0.1',
+              ext_modules=[extension],
+              script_args=["-q", "build"],
+              script_name="C.py",
+              package_dir=build_dir)
+    finally:
+        os.unlink(file_name)
+        os.chdir(home_dir)
+        locker.release()
+
+    # import and return the module
+    lib = "lib.%s-%s" % (get_platform(), sys.version[0:3])    
+    sys.path.append(os.path.join(build_dir, 'build', lib))
+    try:
+        fp, pathname, description = imp.find_module(module_name)
+        module = imp.load_module(module_name, fp, pathname, description)
+    finally:
+        shutil.rmtree(build_dir)
+        if fp:
+            fp.close()
+    return module
 
 
 class C99Handler(object):
     sep = ' ' * 4
     special_functions = {
-        'REFD':lambda args: '(*(%s))' % ', '.join(args), 
-        'ADDR':lambda args: '(&(%s))' % ', '.join(args),
-        'CAST':lambda args: '(%s)(%s)' % (C99Handler.make_type(args[0]), args[1])
-        }
-    substitutions = {'NULL':'NULL', 'True':'1', 'False':'0'}
-    
+        'REFD': lambda args: '(*(%s))' % ', '.join(args),
+        'ADDR': lambda args: '(&(%s))' % ', '.join(args),
+        'CAST': lambda args: '(%s)(%s)' % (C99Handler.make_type(args[0]), args[1])
+    }
+    substitutions = {'NULL': 'NULL', 'True': '1', 'False': '0'}
+
+    def make_types(self, types):
+        for key, value in types.iteritems():
+            types[key] = self.make_type(value)
+        return types
+
     @staticmethod
     def make_type(name):
         newname = stars = ''
@@ -44,6 +105,8 @@ class C99Handler(object):
             newname, name = newname + '__local ', name[6:]
         if name.startswith('const:'):
             newname, name = newname + 'const ', name[6:]
+        if name.startswith('usigned'):
+            newname, name = newname + 'unsigned ', name[6:]
         while name.startswith('ptr_'):
             stars, name = stars + '*', name[4:]
         return newname + name + stars
@@ -69,13 +132,13 @@ class C99Handler(object):
 
     def is_TryExcept(self, item, pad):
         raise NotImplementedError
-        
+
     def is_Break(self, item, pad):
         return 'break;'
-    
+
     def is_Continue(self, item, pad):
         return 'continue;'
-    
+
     def is_While(self, item, pad):
         if item.orelse:
             raise NotImplementedError
@@ -163,19 +226,21 @@ class C99Handler(object):
             raise NotImplementedError
         left, right = item.targets[0], item.value
         if isinstance(left, ast.Name) and not left.id in self.symbols:
-            if isinstance(right, ast.Call) and right.func.id.startswith('new_'):
+            if isinstance(right, ast.Call) and \
+                    right.func.id.startswith('new_'):
                 self.symbols[left.id] = right.func.id[4:]
                 right = right.args[0] if right.args else None
             else:
                 raise RuntimeError('unkown C-type %s' % left.id)
-        return '%s = %s;' % (self.t(item.targets[0]), self.t(right)) if right else ''
+        return '%s = %s;' % (
+            self.t(item.targets[0]), self.t(right)) if right else ''
 
     def is_Call(self, item, pad):
         func = self.t(item.func)
         args = [self.t(a) for a in item.args]
         if func in self.special_functions:
             return self.special_functions[func](args)
-        return '%s(%s)' % (func,args)
+        return '%s(%s)' % (func, ','.join(args))
 
     def is_List(self, item, pad):
         return '{%s}' % ', '.join(self.t(k, pad) for k in item.elts)
@@ -238,11 +303,11 @@ class C99Handler(object):
             if key.startswith('is_'):
                 self.actions[getattr(ast, key[3:])] = getattr(self, key)
 
-    def compile(self, item, types, prefix=''):
+    def convert(self, item, types, prefix=''):
         self.rettype = None  # the return type of the function None is void
         self.symbols = {}
         code = self.is_FunctionDef(item, '', types)
-        vars = ''.join('%s%s %s;\n' % (self.sep, self.make_type(v), k)
+        vars = ''.join('%s%s %s;\n' % (self.sep, v, k)
                        for k, v in self.symbols.items())
         code = code.replace('/*@VARS*/', vars)
         if self.rettype is not None:
@@ -255,10 +320,10 @@ class C99Handler(object):
 
 class JavaScriptHandler(C99Handler):
     special_functions = {'new': lambda args: 'new %s' % ', '.join(args)}
-    substitutions = {'NULL':'null', 'True':'true', 'False':'false'}
+    substitutions = {'NULL': 'null', 'True': 'true', 'False': 'false'}
 
     @staticmethod
-    def make_type(name):
+    def make_type(self, name):
         return name
 
     def is_List(self, item, pad):
@@ -305,7 +370,7 @@ class JavaScriptHandler(C99Handler):
     def __init__(self):
         C99Handler.__init__(self)
 
-    def compile(self, item, types, prefix=''):
+    def convert(self, item, types, prefix=''):
         self.rettype = None  # the return type of the function None is void
         self.symbols = {}
         code = self.is_FunctionDef(item, '')
@@ -315,20 +380,15 @@ class JavaScriptHandler(C99Handler):
         return self.actions[type(item)](item, pad)
 
 
-def ezpy(code, name):
-    ezc = ezpyinline.C(code)
-    return getattr(ezc, name)
-
-
 class Compiler(object):
-    def __init__(self, handler=C99Handler(), filter=None):
+    def __init__(self, handler=None):
         self.functions = {}
-        self.handler = handler
-        self.filter = filter
+        self.handler = handler or C99Handler()
 
     def __call__(self, prefix='', name=None, **types):
         if prefix == 'kernel':
             prefix = '__kernel '
+        types = self.handler.make_types(types)
 
         def wrap(func, types=types, name=name, prefix=prefix):
             if name is None:
@@ -337,19 +397,23 @@ class Compiler(object):
             self.functions[name] = dict(func=func,
                                         prefix=prefix,
                                         ast=decompiled,
-                                        types=types)
-            if self.filter:
-                code = self.handler.compile(decompiled, types, prefix)
-                return self.filter(code, func.__name__)
+                                        types=types,
+                                        code=None)
             return func
         return wrap
 
     def getcode(self, headers=False, constants=None, call=False):
+        """
+        Returns all decorated Python code converted to target language
+        constants is a dict() of constants to be used in convesion
+        headers = True creates the headers (in c99 case only)
+        call = 'f' calls the function f (for JS only)
+        """
         if constants:
             self.handler.constants.update(constants)
         defs, funcs = [], []
         for name, info in self.functions.iteritems():
-            code = self.handler.compile(info['ast'],
+            code = self.handler.convert(info['ast'],
                                         info['types'],
                                         info['prefix'])
             info['code'] = code
@@ -361,6 +425,43 @@ class Compiler(object):
             code = code + '\n\n%s();' % call
         return code
 
+    def compile(self, constants=None):
+        """
+        Compiles all decotared code and returns an
+        returns modules conatining compiled versions of those functions
+        Attention: this function performs a temporary change of directory
+        """
+        map_types = {'unsigned': 'i', 'unsigned int': 'i', 'int': 'i',
+                     'long': 'l', 'float': 'f', 'double': 'd', 'char': 'c',
+                     'short': 'h', 'char*': 's', 'PyObject*': 'O'}
+        filename = 'mdpcl'
+        code = self.getcode(headers=True, constants=constants)
+        python_source = '#include "Python.h"\n\n'
+        python_source += code
+        module_name = 'c' + str(uuid.uuid4()).replace('-', '')
+        funcs = ''
+        for key, info in self.functions.iteritems():
+            func = info['func']
+            name = func.__name__
+            nargs = func.func_code.co_argcount
+            args = ','.join(func.func_code.co_varnames[:nargs])
+            parsing_code = ''
+            for k, v in info['types'].iteritems():
+                parsing_code += C1 % (v, k, map_types[v], k)
+            funcs += C2 % dict(name=name, module_name=module_name)
+            python_source += C3 % dict(module_name=module_name,
+                                       filename=filename,
+                                       name=name, args=args,
+                                       parsing_code=parsing_code)
+        python_source += C4 % dict(module_name=module_name,
+                                   filename=filename,
+                                   name=name, args=args,
+                                   parsing_code=parsing_code,
+                                   funcs=funcs)
+        module = distutil_compile_and_import(module_name, python_source)
+        return module
+
+
 if HAVE_PYOPENCL:
     class Device(object):
         flags = pyopencl.mem_flags
@@ -387,61 +488,67 @@ if HAVE_PYOPENCL:
 
 def test_c99():
     c99 = Compiler()
+
     @c99(a='int', b='int')
     def f(a, b):
         for k in range(n):
             while True:
                 break
-            if k==0 or k!=0 or k<0 or k>0 or not k==0 or \
-               k>=0 or k<=0 or k is None or k is not None:                    
+            if k == 0 or k != 0 or k < 0 or k > 0 or not k == 0 or \
+                    k >= 0 or k <= 0 or k is None or k is not None:
                 continue
-        c = new_int(a+b)
+        c = new_int(a + b)
         printf("%i + %i = %i", a, b, c)
 
-        d = new_ptr_int(CAST(ptr_int,ADDR(c)))
-        c = REFD(d)        
+        d = new_ptr_int(CAST(ptr_int, ADDR(c)))
+        c = REFD(d)
         return c
     print c99.getcode(headers=False, constants=dict(n=10))
 
-def test_EZPY():
-    if not HAVE_EZPYINLINE:
-        logging.error('must install "ezpyinline"')
-        sys.exit(1)
-    c99 = Compiler(filter=ezpy)    
+
+def test_C_compile():
+    c99 = Compiler()
+
     @c99(n='int')
-    def fact(n):
+    def factorial(n):
         output = new_int(1)
-        for k in range(1,n+1):
-            output = output*n
+        for k in range(1, n + 1):
+            output = output * k
         return output
-    print fact(10)
+    compiled = c99.compile()
+    print compiled.factorial(10)
+
 
 def test_OpenCL():
     if not HAVE_PYOPENCL:
         logging.error('must install "pyopencl"')
         sys.exit(1)
     device = Device()
+
     @device.define('kernel',
-               w='global:ptr_float',
-               u='global:const:ptr_float',
-               q='global:const:ptr_float')
-    def solve(w,u,q):
+                   w='global:ptr_float',
+                   u='global:const:ptr_float',
+                   q='global:const:ptr_float')
+    def solve(w, u, q):
         x = new_int(get_global_id(0))
         y = new_int(get_global_id(1))
-        site = new_int(x*n+y)
-        if y!=0 and y!=n-1 and x!=0 and x!=n-1:
-            up = new_int(site-n)
-            down = new_int(site+n)
-            left = new_int(site-1)
-            right = new_int(site+1)
-            w[site] = 1.0/4*(u[up]+u[down]+u[left]+u[right] - q[site])
+        site = new_int(x * n + y)
+        if y != 0 and y != n - 1 and x != 0 and x != n - 1:
+            up = new_int(site - n)
+            down = new_int(site + n)
+            left = new_int(site - 1)
+            right = new_int(site + 1)
+            w[site] = 1.0 / 4 * (u[up] + u[down] + u[left] + u[
+                                 right] - q[site])
     print device.define.getcode(headers=True, constants=dict(n=300))
 
+
 def test_JS():
     js = Compiler(handler=JavaScriptHandler())
+
     @js()
     def f(a):
-        a = new(array(1,2,3,4))
+        a = new(array(1, 2, 3, 4))
         v = [1, 2, 'hello']
         w = {'a': 2, 'b': 4}
 
@@ -455,6 +562,6 @@ def test_JS():
 
 if __name__ == '__main__':
     test_c99()
-    test_EZPY()
+    test_C_compile()
     test_OpenCL()
     test_JS()
